from typing import Optional, List, Union, Any
from pydigitalenergy.models import Extra


class DefaultQos:
    def __init__(self, guid: Optional[str] = '', eRate: Optional[int] = None, inBurst: Optional[int] = None, inRate: Optional[int] = None, **kwargs):
        self.guid = guid
        self.e_rate = eRate
        self.in_rate = inRate
        self.in_burst = inBurst
        self.extra = Extra(kwargs)


class Reservation:
    def __init__(self, vmId: int, ip: str, mac: str, type: str, clientType: Optional[str] = '', desc: Optional[str] = '', domainname: Optional[str] = '', hostname: Optional[str] = '', **kwargs): 
        self.vm_id = vmId
        self.ip = ip
        self.mac = mac
        self.type = type
        self.client_type = clientType
        self.desc = desc
        self.domain_name = domainname
        self.host_name = hostname
        self.extra = Extra(kwargs)


class VNFs:
    def __init__(self, dhcp: Optional[int] = None, **kwargs):
        self.dhcp = dhcp
        self.extra = Extra(kwargs)


class ExternalNetwork:
    def __init__(self, id: int, gid: int, name: str, status: str, ipcidr: str, gateway: str = '', network: str = '', networkId: int = None, checkIps: Optional[List[str]] = [], default: Optional[bool] = None, defaultQos: Optional[Union[DefaultQos, dict]] = None, desc: Optional[str] = '', dns: Optional[List[Any]] = [], excluded: Optional[List[Any]] = [], free_ips: Optional[int] = None, guid: Optional[int] = None, milestones: Optional[int] = None, preReservationsNum: Optional[int] = None, prefix: Optional[int] = None, priVnfDevId: Optional[int] = None, reservations: Optional[List[Reservation]] = [], sharedWith: Optional[List[Any]] = [], vlanId: Optional[int] = None, vnfs: Optional[Union[VNFs, dict]] = None, **kwargs):
        self.id = id
        self.grid_id = gid
        self.name = name
        self.status = status
        self.ip_cidr = ipcidr
        self.gateway = gateway
        self.network = network
        self.network_id = networkId
        self.check_ips = checkIps
        self.default = default
        self.default_qos = DefaultQos(**defaultQos) if isinstance(defaultQos, dict) else defaultQos
        self.description = desc
        self.dns = dns
        self.excluded = excluded
        self.free_ips = free_ips
        self.guid = guid
        self.milestones = milestones
        self.pre_reservations_num = preReservationsNum
        self.prefix = prefix
        self.pri_vnf_dev_id = priVnfDevId
        self.reservations = [] if not reservations else [Reservation(**reservation) for reservation in reservations]
        self.shared_with = sharedWith
        self.vlan_id = vlanId
        self.vnfs = VNFs(**vnfs) if isinstance(vnfs, dict) else vnfs
        self.extra = Extra(kwargs)



# List of API endpoints PyDigitalEnergy Cloudbroker knows about

API_PATH = {
    # Compute
    'compute_list': '/cloudbroker/compute/list',
    'compute_get': '/cloudbroker/compute/get',
    # Stack
    'stack_list': '/cloudbroker/stack/list',
    'stack_get': '/cloudbroker/stack/get',
    # Node
    'node_list': '/cloudbroker/node/list',
    'node_get': '/cloudbroker/node/get',
    # Grid
    'grid_list': '/cloudbroker/grid/list',
    'grid_get': '/cloudbroker/grid/get',
    # Storage
    'sep_list': '/cloudbroker/sep/list',
    'sep_get': '/cloudbroker/sep/get',
    'sep_get_config': '/cloudbroker/sep/getConfig',
    'sep_get_pool': '/cloudbroker/sep/getPool',
    # Image
    'image_list': '/cloudbroker/image/list',
    'image_get': '/cloudbroker/image/get',
    # Disk
    'disks_list': '/cloudbroker/disks/list',
    'disks_get': '/cloudbroker/disks/get',
    # Network
    'extnet_list': '/cloudbroker/extnet/list',
    'extnet_get': '/cloudbroker/extnet/get',
}


from typing import List
from pydigitalenergy.cloudapi import models
from pydigitalenergy.cloudapi.endpoints import API_PATH
from pydigitalenergy.adapter import RestAdapter
from pydigitalenergy.basegroup import BaseGroup


class Cloudapi:
    def __init__(self, adapter: RestAdapter):
        self.virtual_machines = MachineGroup(adapter)


class MachineGroup(BaseGroup):
    """
    Virtual Machine Group
    """

    def list(self, include_deleted: bool = False, cloudspace_id: int = 0, page: int = 0, size: int = 0) -> List[models.Machine]:
        """
        Get list of virtual machine instances
        Page and size can be specified only together

        :param include_deleted: (optional)
            Whether add deleted machine instances or not
        :param cloudspace_id: (optional)
            Cloud space identifier which contains that machine instance
        :param page: (optional)
            Page number
        :param size: (optional)
            Amount of instances per page
        
        :return: 
            List of Machine objects
        """
        result = self._get_data(API_PATH['machine_list'], {'includedeleted': include_deleted, 'cloudspaceId': cloudspace_id, 'page': page, 'size': size})
        return [models.Machine(**machine) for machine in result.data]

    def get(self, machine_id: int) -> models.Machine:
        """
        Get information about virtual machine instance

        :param machine_id: 
            Machine instance identifier
        
        :return: 
            Machine object
        """
        result = self._get_data(API_PATH['machine_get'], {'machineId': machine_id})
        return models.Machine(**result.data)




from typing import Optional, List, Dict, Union, Any
from pydigitalenergy.models import Extra


class DiskDelQueue:
    def __init__(self, purge_attempts_threshold: Optional[int] = None, **kwargs):
        self.purge_attempts_threshold = purge_attempts_threshold
        self.extra = Extra(kwargs)


class HousekeepingSettings:
    def __init__(self, disk_del_queue: Optional[Union[DiskDelQueue, dict]] = None, **kwargs):
        self.disk_del_queue = DiskDelQueue(**disk_del_queue) if isinstance(disk_del_queue, dict) else disk_del_queue
        self.extra = Extra(kwargs)


class StoragePool:
    def __init__(self, id: int, snapshot_pool_id: int, name: str, clone_technology: Optional[str] = '', maxLdevId: Optional[int] = None, minLdevId: Optional[int] = None, snapshotable: Optional[bool] = None, types: Optional[List[str]] = [], usage_limit: Optional[int] = None, **kwargs):
        self.id = id
        self.snapshot_pool_id = snapshot_pool_id
        self.name = name
        self.clone_technology = clone_technology
        self.max_ldev_id = maxLdevId
        self.min_ldev_id = minLdevId
        self.snapshotable = snapshotable
        self.types = types
        self.usage_limit = usage_limit
        self.extra = Extra(kwargs)


class StorageConfig:
    def __init__(self, model: str, protocol: str, ssl_verify: bool, SN: int, disk_max_size: int, API_URLs: Optional[List[str]] = [], format: Optional[str] = '', hostGroupNumMax: Optional[int] = None, hostGroupNumMin: Optional[int] = None, hostGroupNumber: Optional[int] = None, hosts: Optional[Dict[str, int]] = {}, housekeeping_settings: Optional[Union[HousekeepingSettings, dict]] = None, mgmt_password: Optional[str] = '', mgmt_user: Optional[str] = '', name_prefix: Optional[str] = '', pools: Optional[List[Union[StoragePool, dict]]] = [], ports: Optional[List[str]] = [], **kwargs):
        self.model = model
        self.protocol = protocol
        self.ssl_verify = ssl_verify
        self.serial_number = SN
        self.disk_max_size = disk_max_size
        self.api_urls = API_URLs
        self.format = format
        self.host_group_num_max = hostGroupNumMax
        self.host_group_num_min = hostGroupNumMin
        self.host_group_number = hostGroupNumber
        self.hosts = hosts
        self.housekeeping_settings = HousekeepingSettings(**housekeeping_settings) if isinstance(housekeeping_settings, dict) else housekeeping_settings
        self.mgmt_password = mgmt_password
        self.mgmt_user = mgmt_user
        self.name_prefix = name_prefix
        self.pools = [] if not pools else [StoragePool(**pool) for pool in pools]
        self.ports = ports
        self.extra = Extra(kwargs)


class StorageEndpoint:
    def __init__(self, id: int, gid: int, name: str, desc: str, type: str, objStatus: str, techStatus: str, config: Optional[Union[StorageConfig, dict]] = None, consumedBy: Optional[List[int]] = [], guid: Optional[int] = None, milestones: Optional[int] = None, providedBy: Optional[List[Any]] = [], **kwargs):
        self.id = id
        self.grid_id = gid
        self.name = name
        self.description = desc
        self.type = type
        self.obj_status = objStatus
        self.tech_status = techStatus
        self.config = StorageConfig(**config) if isinstance(config, dict) else config
        self.consumed_by = consumedBy
        self.guid = guid
        self.milestones = milestones
        self.provided_by = providedBy
        self.extra = Extra(kwargs)



from typing import Union, Optional, List
from pydigitalenergy.models import Extra


class DockerRegistry:
    def __init__(self, server: str, username: Optional[str] = '', password: Optional[str] = '', **kwargs):
        self.server = server
        self.username = username
        self.password = password
        self.extra = Extra(kwargs)


class Email:
    def __init__(self, enabled: bool, address: Optional[str] = '', **kwargs):
        self.enabled = enabled
        self.address = address
        self.extra = Extra(kwargs)


class HealthCheckNotification:
    def __init__(self, emails: Optional[List[Email]] = [], **kwargs):
        self.emails = [] if not emails else [Email(**email) for email in emails]
        self.extra = Extra(kwargs)


class Net:
    def __init__(self, eRate: Optional[int] = None, inBurst: Optional[int] = None, inRate: Optional[int] = None, **kwargs):
        self.e_rate = eRate
        self.in_burst = inBurst
        self.in_rate = inRate
        self.extra = Extra(kwargs)


class NetQos:
    def __init__(self, ext_net: Union[Net, dict], vins: Union[Net, dict], **kwargs):
        self.ext_net = Net(**ext_net) if isinstance(ext_net, dict) else ext_net
        self.vins = Net(**vins) if isinstance(vins, dict) else vins
        self.extra = Extra(kwargs)


class Prometheus:
    def __init__(self, scrapeInterval: int, **kwargs):
        self.scrape_interval = scrapeInterval
        self.extra = Extra(kwargs)


class Settings:
    def __init__(self, enableUptimeMonitor: bool, allowedports: Optional[List[int]] = [], cleanupRetentionPeriod: Optional[int] = None, docker_registry: Optional[DockerRegistry] = None, extnetMaxPreReservationsNum: Optional[int] = None, healthcheck_notifications: Optional[HealthCheckNotification] = None, limits = None, net_qos: Optional[NetQos] = None, prometheus: Optional[Prometheus] = None, registrationPassword: Optional[str] = '', registrationUrl: Optional[str] = '', registrationUser: Optional[str] = '', vinsMaxPreReservationsNum: Optional[int] = None, vnfdev_mgmt_net_range: Optional[str] = '', **kwargs):
        self.enable_uptime_monitor = enableUptimeMonitor
        self.allowed_ports = allowedports
        self.cleanup_retention_period = cleanupRetentionPeriod
        self.docker_registry = DockerRegistry(**docker_registry) if isinstance(docker_registry, dict) else docker_registry
        self.extnet_max_pre_reservations_num = extnetMaxPreReservationsNum
        self.healthcheck_notifications = HealthCheckNotification(**healthcheck_notifications) if isinstance(healthcheck_notifications, dict) else healthcheck_notifications
        self.limits = limits
        self.net_qos = NetQos(**net_qos) if isinstance(net_qos, dict) else net_qos
        self.prometheus = Prometheus(**prometheus) if isinstance(prometheus, dict) else prometheus
        self.registration_assword = registrationPassword
        self.registration_url = registrationUrl
        self.registration_ser = registrationUser
        self.vins_max_pre_reservations_num = vinsMaxPreReservationsNum
        self.vnfdev_mgmt_net_range = vnfdev_mgmt_net_range
        self.extra = Extra(kwargs)


class Grid:
    def __init__(self, id: int, name: str, guid: Optional[int] = None, nid: Optional[int] = None, settings: Optional[Settings] = None, useavahi: Optional[int] = None, **kwargs):
        self.id = id
        self.name = name
        self.guid = guid
        self.nid = nid
        self.settings = Settings(**settings) if isinstance(settings, dict) else settings
        self.useavahi = useavahi
        self.extra = Extra(kwargs)



from typing import Optional, List, Union
from pydigitalenergy.models import Extra


class Stack:
    def __init__(self, id: int, gid: int, name: str, status: str, type: str, referenceId: int, apiUrl: Optional[str] = '', apikey: Optional[str] = '', appId: Optional[str] = '', descr: Optional[str] = '', drivers: Optional[List[str]] = '', eco: Optional[str] = '', error: Optional[int] = None, guid: Optional[int] = None, images: Optional[List[int]] = [], **kwargs):
        self.id = id
        self.grid_id = gid
        self.name = name
        self.status = status
        self.type = type
        self.reference_id = int(referenceId) if referenceId.isdigit() else referenceId
        self.api_url = apiUrl
        self.api_key = apikey
        self.app_id = appId
        self.description = descr
        self.drivers = drivers
        self.eco = eco
        self.error = error
        self.guid = guid
        self.images = images
        self.extra = Extra(kwargs)



from uuid import UUID
from typing import Optional, Union, List, Any
from pydigitalenergy.cloudbroker.models.compute import Affinity, Boot, Disk, Interface, OSUser, SnapSet, AntiAffinityRule
from pydigitalenergy.cloudbroker.models.host import Host, Consumption, StorageConfig
from pydigitalenergy.cloudbroker.models.cluster import Cluster
from pydigitalenergy.cloudbroker.models.image import Image
from pydigitalenergy.models import Extra


class VM:
    def __init__(self, 
        # Required
        id: int, name: str, gid: int, accountId: int, accountName: str, rgId: int, rgName: str, status: str, techStatus: str, 
        hostId: int, hostGid: int, hostName: str, hostStatus: str, hostType: str, hostStackId: int, hostSepName: str, hostDesc: str, hostSepType: str, hostObjStatus: str, hostTechStatus: str, hostReferenceId: int,
        clusterId: int, clusterGid: int, clusterName: str, clusterStatus: str, clusterReferenceId: int, 
        # Optional
        affinityLabel: Optional[str] = '', affinityRules: Optional[List[Any]] = [], affinityWeight: Optional[int] = None, antiAffinityRules: Optional[List[Union[AntiAffinityRule, dict]]] = [], arch: Optional[str] = '', bootOrder: Optional[List[str]] = [], bootdiskSize: Optional[int] = None, cloneReference: Optional[int] = None, clones: Optional[List[Any]] = [], computeciId: Optional[int] = None, cpus: Optional[int] = None, createdBy: Optional[str] = '', createdTime: Optional[int] = None, customFields: Optional[Any] = None, deletedBy: Optional[str] = '', deletedTime: Optional[int] = None, desc: Optional[str] = '', devices: Optional[Any] = None, disks: Optional[List[Union[Disk, dict, int]]] = [], driver: Optional[str] = '', guid: Optional[int] = None, interfaces: Optional[List[Union[Interface, dict]]] = [], lockStatus: Optional[str] = '', managerId: Optional[int] = None, managerType: Optional[str] = '', migrationjob: Optional[int] = None, milestones: Optional[int] = None, osUsers: Optional[List[Union[OSUser, dict]]] = [], pinned: Optional[bool] = None, ram: Optional[int] = None, referenceId: Optional[UUID] = None, registered: Optional[bool] = None, resName: Optional[str] = '', snapSets: Optional[List[Union[SnapSet, dict]]] = [], statelessSepId: Optional[int] = None, statelessSepType: Optional[str] = '', tags: Optional[Any] = None, totalDisksSize: Optional[int] = None, updatedBy: Optional[str] = '', updatedTime: Optional[int] = None, userManaged: Optional[bool] = None, userdata: Optional[Any] = None, vgpus: Optional[List[Any]] = [], vinsConnected: Optional[int] = None, virtualImageId: Optional[int] = None,
        hostVersion: Optional[str] = '', hostApiUrl: Optional[str] = '', hostApikey: Optional[str] = '', hostDescr: Optional[str] = '', hostDrivers: Optional[List[str]] = '', hostError: Optional[int] = None, hostGuid: Optional[int] = None, hostIpaddr: Optional[List[str]] = [], hostConsumption: Union[Consumption, dict] = None, hostRoles: Optional[List[str]] = [], hostConfig: Optional[Union[StorageConfig, dict]] = None, hostConsumedBy: Optional[List[int]] = [], hostMilestones: Optional[int] = None, hostProvidedBy: Optional[List[Any]] = [],
        clusterDescr: Optional[str] = '', clusterError: Optional[int] = None, clusterGuid: Optional[int] = None, clusterConsumption: Union[Consumption, dict] = None, clusterRoles: Optional[List[str]] = [],
        imageHasImage: bool = True, imageId: int = None, imageName: str = '', imageType: str = '', imageVersion: str = '', imageStatus: str = '', imageTechStatus: str = '', imageUNCPath: Optional[str] = '', imageAccountId: Optional[int] = None, imageArchitecture: Optional[str] = '', imageBootType: Optional[str] = '', imageBootable: Optional[bool] = None, imageComputeciId: Optional[int] = None, imageDeletedTime: Optional[int] = None, imageDesc: Optional[str] = '', imageDrivers: Optional[List[str]] = [], imageEnabled: Optional[bool] = None, imageGid: Optional[int] = None, imageGuid: Optional[int] = None, imageHistory: Optional[List[Any]] = [], imageHotResize: Optional[bool] = None, imageLastModified: Optional[int] = None, imageLinkTo: Optional[int] = None, imageMilestones: Optional[int] = None, imageUsername: Optional[str] = '', imagePool: Optional[str] = '', imageProvidername: Optional[str] = '', imagePurgeAttempts: Optional[int] = None, imageReferenceId: Optional[str] = '', imageResId: Optional[int] = None, imageResName: Optional[str] = '', imageRescuecd: Optional[bool] = None, imageSepId: Optional[int] = None, imageSharedWith: Optional[List[Any]] = [], imageSize: Optional[int] = None, imageUrl: Optional[str] = '', imageVirtual: Optional[bool] = None,
     **kwargs):
        self.id = id
        self.data_center_id = gid
        self.account_id = accountId
        self.resource_group_id = rgId
        self.name = name
        self.account_name = accountName
        self.resource_group_name = rgName
        self.status = status
        self.tech_status = techStatus
        self.affinity = Affinity(affinityLabel, affinityRules, affinityWeight, antiAffinityRules)
        self.architecture = arch
        self.boot = Boot(bootOrder, bootdiskSize)
        self.clone_reference = cloneReference
        self.clones = clones
        self.compute_ci_id = computeciId
        self.cpus = cpus
        self.created_by = createdBy
        self.created_time = createdTime
        self.deleted_by = deletedBy
        self.deleted_time = deletedTime
        self.custom_fields = customFields
        self.description = desc
        self.devices = devices
        self.disks = [] if not disks else [Disk(**disk) if isinstance(disk, dict) else disk for disk in disks]
        self.driver = driver
        self.guid = guid
        self.interfaces = [] if not interfaces else [Interface(**interface) if isinstance(interface, dict) else interface for interface in interfaces]
        self.lock_status = lockStatus
        self.manager_id = managerId
        self.manager_type = managerType
        self.migration_job = migrationjob
        self.milestones = milestones
        self.os_users = [] if not osUsers else [OSUser(**user) if isinstance(user, dict) else user for user in osUsers]
        self.pinned = pinned
        self.ram = ram
        self.reference_id = referenceId
        self.registered = registered
        self.res_name = resName
        self.snap_sets = [] if not snapSets else [SnapSet(**snap_set) if isinstance(snap_set, dict) else snap_set for snap_set in snapSets]
        self.stateless_sep_id = statelessSepId
        self.stateless_sep_type = statelessSepType
        self.tags = tags
        self.total_disks_size = totalDisksSize
        self.updated_by = updatedBy
        self.updated_time = updatedTime
        self.user_managed = userManaged
        self.user_data = userdata
        self.vgpus = vgpus
        self.vins_connected = vinsConnected
        self.virtual_image_id = virtualImageId
        self.has_image = imageHasImage
        self.host = Host(hostId, hostGid, hostName, hostStatus, hostType, hostStackId, hostSepName, hostDesc, hostSepType, hostObjStatus, hostTechStatus, hostVersion, hostReferenceId, hostApiUrl, hostApikey, hostDescr, hostDrivers, hostError, hostGuid, hostIpaddr, hostConsumption, hostRoles, hostConfig, hostConsumedBy, hostMilestones, hostProvidedBy)
        self.cluster = Cluster(clusterId, clusterGid, clusterName, clusterStatus, clusterReferenceId, clusterDescr, clusterError, clusterGuid, clusterConsumption, clusterRoles)
        self.image = Image(imageId, imageName, imageType, imageVersion, imageStatus, imageTechStatus, imageUNCPath, imageAccountId, imageArchitecture, imageBootType, imageBootable, imageComputeciId, imageDeletedTime, imageDesc, imageDrivers, imageEnabled, imageGid, imageGuid, imageHistory, imageHotResize, imageLastModified, imageLinkTo, imageMilestones, imageUsername, imagePool, imageProvidername, imagePurgeAttempts, imageReferenceId, imageResId, imageResName, imageRescuecd, imageSepId, imageSharedWith, imageSize, imageUrl, imageVirtual)
        self.extra = Extra(kwargs)



from unittest import TestCase
from unittest.mock import MagicMock
from pydigitalenergy.models import Result
from pydigitalenergy.cloudbroker import models
from pydigitalenergy.digitalenergy import DigitalEnergyApi


class TestCloudbrokerApi(TestCase):

    required_test_settings = {
        'hostname': 'localhost',
        'client_id': 'test',
        'client_secret': 'test',
        'token': 'test-jwt-key',
    } 

    def setUp(self) -> None:
        self.api = DigitalEnergyApi(**self.required_test_settings)
        self.api._adapter.get = MagicMock()
        self.api._adapter.post = MagicMock()

    data_compute = {
        'id': 1, 
        'name': 'vm', 
        'gid': 1, 
        'accountId': 1, 
        'accountName': 'test', 
        'rgId': 1,
        'rgName': 'test', 
        'stackId': 1, 
        'status': 'enabled',
        'techStatus': 'started'
    }

    def test_computes_get_returns_one_compute(self):
        self.api._adapter.post.return_value = Result(200, headers={}, data=self.data_compute)
        compute = self.api.cloudbroker.computes.get(1)
        self.assertIsInstance(compute, models.Compute)

    def test_computes_list_returns_list_of_compute(self):
        self.api._adapter.post.return_value = Result(200, headers={}, data=[self.data_compute for i in range(2)])
        compute_list = self.api.cloudbroker.computes.list()
        self.assertIsInstance(compute_list, list)
        self.assertTrue(len(compute_list), 2)
        self.assertIsInstance(compute_list[0], models.Compute)



from typing import Optional, List, Union, Any
from pydigitalenergy.models import Extra


class Image:
    def __init__(self, id: int, name: str, type: str, version: str, status: str, techStatus: str, UNCPath: Optional[str] = '', accountId: Optional[int] = None, architecture: Optional[str] = '', bootType: Optional[str] = '', bootable: Optional[bool] = None, computeciId: Optional[int] = None, deletedTime: Optional[int] = None, desc: Optional[str] = '', drivers: Optional[List[str]] = [], enabled: Optional[bool] = None, gid: Optional[int] = None, guid: Optional[int] = None, history: Optional[List[Any]] = [], hotResize: Optional[bool] = None, lastModified: Optional[int] = None, linkTo: Optional[int] = None, milestones: Optional[int] = None, username: Optional[str] = '', pool: Optional[str] = '', providername: Optional[str] = '', purgeAttempts: Optional[int] = None, referenceId: Optional[str] = '', resId: Optional[int] = None, resName: Optional[str] = '', rescuecd: Optional[bool] = None, sepId: Optional[int] = None, sharedWith: Optional[List[Any]] = [], size: Optional[int] = None, url: Optional[str] = '', virtual: Optional[bool] = None, **kwargs):
        self.id = id
        self.name = name
        self.type = type
        self.version = version
        self.status = status
        self.tech_status = techStatus
        self.unc_patch = UNCPath
        self.account_id = accountId
        self.architecture = architecture
        self.boot_type = bootType
        self.bootable = bootable
        self.compute_ci_id = computeciId
        self.deleted_time = deletedTime
        self.description = desc
        self.drivers = drivers
        self.enabled = enabled
        self.grid_id = gid
        self.guid = guid
        self.history = history
        self.hot_resize = hotResize
        self.last_modified = lastModified
        self.link_to = linkTo
        self.milestones = milestones
        self.pool = pool
        self.provider_name = providername
        self.purge_attempts = purgeAttempts
        self.reference_id = referenceId
        self.res_id = resId
        self.res_name = resName
        self.rescuecd = rescuecd
        self.sep_id = sepId
        self.shared_with = sharedWith
        self.size = size
        self.url = url
        self.username = username
        self.virtual = virtual
        self.extra = Extra(kwargs)



from logging import Logger
from pydigitalenergy.adapter import RestAdapter
from pydigitalenergy.system.actors import System
from pydigitalenergy.cloudapi.actors import Cloudapi
from pydigitalenergy.cloudbroker.actors import Cloudbroker


class DigitalEnergyApi:
    def __init__(self, hostname: str, client_id: str, client_secret: str, auth_version: str = 'v1', auth_validity: int = 3600, ssl_verify: bool = True, logger: Logger = None, token: str = ''):
        """
        The DigitalEnergyApi class provides convenient access to Digital Energy's API

        :param hostname: 
            DigitalEnergy API hostname. Something like, api.digitalenergy.online
        :param client_id: 
            String defined application identifier
        :param client_secret: 
            String defined application secret
        :param auth_version: 
            Auth version, default v1
        :param auth_validity: 
            Lifetime of JWT-key in seconds, default 3600 seconds or 1 hour
        :param ssl_verify: 
            Normally set to True, but if having SSL/TLS cert validation issues, can turn off with False
        :param logger: (optional)
            If your app has a logger, pass it in here
        :param token: (optional)
            Your JWT-key to work with API, if not specified will be generated automatically
        """
        self._adapter = RestAdapter(hostname, client_id, client_secret, auth_version, auth_validity, ssl_verify, logger, token)
        self._logger = logger
        
        # Following actors are RESTfull API interfaces that are used to run backend tasks
        self.system = System(self._adapter)
        # self.cloudapi = Cloudapi(self._adapter)
        self.cloudbroker = Cloudbroker(self._adapter)
        # self.libcloud = None



from uuid import UUID
from typing import Union, Optional, List, Any
from pydigitalenergy.cloudbroker.models.disk import Disk
from pydigitalenergy.models import Extra


class AntiAffinityRule:
    def __init__(self, key: str, guid: Optional[str] = '', mode: Optional[str] = '', value: Optional[bool] = None, **kwargs):
        self.key = key
        self.guid = guid
        self.mode = mode
        self.value = value
        self.extra = Extra(kwargs)


class Interface:
    def __init__(self, connId: int, connType: str, defGw: Optional[str] = '', flipgroupId: Optional[int] = None, guid: Optional[str] = '', ipAddress: Optional[str] = '', listenSsh: Optional[bool] = None, mac: Optional[str] = '', name: Optional[str] = '', netId: Optional[int] = None, netMask: Optional[int] = None, netType: Optional[str] = '', pciSlot: Optional[int] = None, target: Optional[str] = '', type: Optional[str] = '', vnfs: Optional[List[Any]] = [], **kwargs):
        self.conn_id = connId
        self.conn_type = connType
        self.def_gw = defGw
        self.flipgroup_id = flipgroupId
        self.guid = guid
        self.ip_address = ipAddress
        self.listen_ssh = listenSsh
        self.mac = mac
        self.name = name
        self.net_id = netId
        self.net_mask = netMask
        self.net_type = netType
        self.pci_slot = pciSlot
        self.target = target
        self.type = type
        self.vnfs = vnfs
        self.extra = Extra(kwargs)


class OSUser:
    def __init__(self, guid: Optional[str] = '', login: Optional[str] = '', password: Optional[str] = '', pubkey: Optional[str] = '', **kwargs):
        self.guid = guid
        self.login = login
        self.password = password
        self.pubkey = pubkey
        self.extra = Extra(kwargs)


class SnapSet:
    def __init__(self, guid: Optional[UUID] = '', disks: Optional[List[int]] = [], label: Optional[str] = '', timestamp: Optional[int] = None, **kwargs):
        self.guid = guid
        self.disks = disks
        self.label = label
        self.timestamp = timestamp
        self.extra = Extra(kwargs)


class Affinity:
    def __init__(self, affinityLabel: Optional[str] = '', affinityRules: Optional[List[Any]] = [], affinityWeight: Optional[int] = None, antiAffinityRules: Optional[List[Union[AntiAffinityRule, dict]]] = []):
        self.affinity_label = affinityLabel
        self.affinity_rules = affinityRules
        self.affinity_weight = affinityWeight
        self.anti_affinity_rules = [] if not antiAffinityRules else [AntiAffinityRule(**rules) for rules in antiAffinityRules]


class Boot:
    def __init__(self, bootOrder: Optional[List[str]] = [], bootdiskSize: Optional[int] = None):
        self.boot_order = bootOrder
        self.boot_disk_size = bootdiskSize


class Compute:
    def __init__(self, id: int, name: str, gid: int, accountId: int, accountName: str, rgId: int, rgName: str, stackId: int, status: str, techStatus: str, affinityLabel: Optional[str] = '', affinityRules: Optional[List[Any]] = [], affinityWeight: Optional[int] = None, antiAffinityRules: Optional[List[Union[AntiAffinityRule, dict]]] = [], arch: Optional[str] = '', bootOrder: Optional[List[str]] = [], bootdiskSize: Optional[int] = None, cloneReference: Optional[int] = None, clones: Optional[List[Any]] = [], computeciId: Optional[int] = None, cpus: Optional[int] = None, createdBy: Optional[str] = '', createdTime: Optional[int] = None, customFields: Optional[Any] = None, deletedBy: Optional[str] = '', deletedTime: Optional[int] = None, desc: Optional[str] = '', devices: Optional[Any] = None, disks: Optional[List[Union[Disk, dict, int]]] = [], driver: Optional[str] = '', guid: Optional[int] = None, imageId: Optional[int] = None, interfaces: Optional[List[Union[Interface, dict]]] = [], lockStatus: Optional[str] = '', managerId: Optional[int] = None, managerType: Optional[str] = '', migrationjob: Optional[int] = None, milestones: Optional[int] = None, osUsers: Optional[List[Union[OSUser, dict]]] = [], pinned: Optional[bool] = None, ram: Optional[int] = None, referenceId: Optional[UUID] = None, registered: Optional[bool] = None, resName: Optional[str] = '', snapSets: Optional[List[Union[SnapSet, dict]]] = [], statelessSepId: Optional[int] = None, statelessSepType: Optional[str] = '', tags: Optional[Any] = None, totalDisksSize: Optional[int] = None, updatedBy: Optional[str] = '', updatedTime: Optional[int] = None, userManaged: Optional[bool] = None, userdata: Optional[Any] = None, vgpus: Optional[List[Any]] = [], vinsConnected: Optional[int] = None, virtualImageId: Optional[int] = None, **kwargs):
        self.id = id
        self.grid_id = gid
        self.stack_id = stackId
        self.account_id = accountId
        self.resource_group_id = rgId
        self.image_id = imageId
        self.name = name
        self.account_name = accountName
        self.resource_group_name = rgName
        self.status = status
        self.tech_status = techStatus
        self.affinity = Affinity(affinityLabel, affinityRules, affinityWeight, antiAffinityRules)
        self.architecture = arch
        self.boot = Boot(bootOrder, bootdiskSize)
        self.clone_reference = cloneReference
        self.clones = clones
        self.compute_ci_id = computeciId
        self.cpus = cpus
        self.created_by = createdBy
        self.created_time = createdTime
        self.deleted_by = deletedBy
        self.deleted_time = deletedTime
        self.custom_fields = customFields
        self.description = desc
        self.devices = devices
        self.disks = [] if not disks else [Disk(**disk) if isinstance(disk, dict) else disk for disk in disks]
        self.driver = driver
        self.guid = guid
        self.interfaces = [] if not interfaces else [Interface(**interface) if isinstance(interface, dict) else interface for interface in interfaces]
        self.lock_status = lockStatus
        self.manager_id = managerId
        self.manager_type = managerType
        self.migration_job = migrationjob
        self.milestones = milestones
        self.os_users = [] if not osUsers else [OSUser(**user) if isinstance(user, dict) else user for user in osUsers]
        self.pinned = pinned
        self.ram = ram
        self.reference_id = referenceId
        self.registered = registered
        self.res_name = resName
        self.snap_sets = [] if not snapSets else [SnapSet(**snap_set) if isinstance(snap_set, dict) else snap_set for snap_set in snapSets]
        self.stateless_sep_id = statelessSepId
        self.stateless_sep_type = statelessSepType
        self.tags = tags
        self.total_disks_size = totalDisksSize
        self.updated_by = updatedBy
        self.updated_time = updatedTime
        self.user_managed = userManaged
        self.user_data = userdata
        self.vgpus = vgpus
        self.vins_connected = vinsConnected
        self.virtual_image_id = virtualImageId
        self.extra = Extra(kwargs)



from copy import deepcopy
from typing import Union, Tuple, Dict, List
from pydigitalenergy.models import Result
from pydigitalenergy.adapter import RestAdapter


class BaseGroup:
    def __init__(self, adapter: RestAdapter):
        """
        Base class for all Digital Energy API Groups

        :param adapter: 
            Low-lever adapter for Digital Energy API
        """
        self._adapter = adapter

    def _get_data(self, endpoint, ep_params = None, method='POST') -> Result:
        """
        Get raw data using REST adapter methods post or get

        :param endpoint: 
            URL Endpoint as a string
        :param ep_params: (Optional)
            Dictionary of Endpoint parameters
        :param method: 
            HTTP method in this case POST or GET
        """
        if method.upper() == 'POST':
            return self._adapter.post(endpoint, ep_params)
        elif method.upper() == 'GET':
            return self._adapter.get(endpoint, ep_params)

    def _remove_fields(self, data: Dict, fields: Union[List, Tuple]):
        """ 
        Remove redundant fields from recieved data

        :param data:
            Dictionary with data recieved via API
        :param fields:
            List or tuple with fields which should be removed from data
        """
        for key in fields:
            data.pop(key, None)

    def _rename_fields(self, data: Dict, names: Dict):
        """ 
        Remove redundant fields from recieved data

        :param data:
            Dictionary with data recieved via API
        :param names:
            Dictionary with the followind structure {'old_name': 'new_name'}
        """
        for old_name, new_name in names.items():
            if old_name in data:
                data[new_name] = data.pop(old_name)

    def _add_fields_prefix(self, data: Dict, prefix: str):
        """ 
        Add prefix to all recieved data fields 

        :param data:
            Dictionary with data recieved via API
        :param prefix:
            Prefix which should be added before every field name
        """
        source_data = deepcopy(data)
        for key, val in source_data.items():
            data[prefix + key[0].upper() + key[1:]] = data.pop(key)



from typing import List
from pydigitalenergy.system import models
from pydigitalenergy.system.endpoints import API_PATH
from pydigitalenergy.adapter import RestAdapter
from pydigitalenergy.basegroup import BaseGroup
from pydigitalenergy.exceptions import DigitalEnergyApiException


class System:
    def __init__(self, adapter: RestAdapter):
        self.grids = GridGroup(adapter)
        self.data_centers = DataCenterGroup(adapter)


class GridGroup(BaseGroup):
    """
    System Grid Group
    """

    def list(self) -> List[models.Grid]:
        """
        Get list of system grid instances

        :return: 
            List of Grid objects
        """
        result = self._get_data(API_PATH['grid_list'], method='GET')
        return [models.Grid(**grid) for grid in result.data]

    def get(self, grid_id: int) -> models.Grid:
        """
        Get information about system grid instance

        :param grid_id: 
            Grid instance identifier

        :return: 
            Grid object
        """
        for grid in self.list():
            if grid.id == grid_id:
                return grid
        raise DigitalEnergyApiException('404: Not Found')


class DataCenterGroup(GridGroup):
    """
    System Data Center Group 
    It is an alias for System Grid Group
    """
    pass


from typing import Union, Optional, List
from pydigitalenergy.models import Extra


class Consumed:
    def __init__(self, RAM: int, computes: int, routers: int, vCPU: int, **kwargs):
        self.ram = RAM
        self.computes = computes
        self.routers = routers
        self.vcpu = vCPU
        self.extra = Extra(kwargs)


class Storage:
    def __init__(self, RAM: int, **kwargs):
        self.ram = RAM
        self.extra = Extra(kwargs)


class Consumption:
    def __init__(self, consumed: Union[Consumed, dict], free: Union[Storage, dict], hostname: str, reserved: Union[Storage, dict], total: Union[Storage, dict], **kwargs):
        self.consumed = Consumed(**consumed) if isinstance(consumed, dict) else consumed
        self.free = Storage(**free) if isinstance(free, dict) else free
        self.hostname = hostname
        self.reserved = Storage(**reserved) if isinstance(reserved, dict) else reserved
        self.total = Storage(**total) if isinstance(total, dict) else total
        self.extra = Extra(kwargs)


class Node:
    def __init__(self, gid: int, id: int, name: str, status: Optional[str], version: Optional[str], ipaddr: Optional[List[str]] = [], consumption: Union[Consumption, dict] = None, roles: Optional[List[str]] = [], **kwargs):
        self.id = id
        self.grid_id = gid
        self.name = name
        self.status = status
        self.version = version
        self.ip_address = ipaddr
        self.consumption = Consumption(**consumption) if isinstance(consumption, dict) else consumption
        self.roles = roles
        self.extra = Extra(kwargs)



from typing import Union, Optional, List, Any
from pydigitalenergy.models import Extra


class IOtune:
    def __init__(self, total_iops_sec: Optional[int], **kwargs):
        self.total_iops_sec = total_iops_sec
        self.extra = Extra(kwargs)


class Disk:
    def __init__(self, id: int, gid: int, name: str, accountId: Optional[int] = None, accountName: Optional[str] = '', bootPartition: Optional[int] = None, createdTime: Optional[int] = None, deletedTime: Optional[int] = None, desc: Optional[str] = '', destructionTime: Optional[int] = None, devicename: Optional[str] = '', diskPath: Optional[str] = '', guid: Optional[int] = None, imageId: Optional[int] = None, images: Optional[List[Any]] = [], iotune: Optional[Union[IOtune, dict]] = None, milestones: Optional[int] = None, order: Optional[int] = None, params: Optional[str] = '', parentId: Optional[int] = None, pciSlot: Optional[int] = None, pool: Optional[str] = '', purgeAttempts: Optional[int] = None, purgeTime: Optional[int] = None, realityDeviceNumber: Optional[int] = None, referenceId: Optional[str] = '', resId: Optional[int] = None, resName: Optional[str] = '', role: Optional[str] = '', sepId: Optional[int] = None, sepType: Optional[str] = '', sizeMax: Optional[int] = None, sizeUsed: Optional[int] = None, snapshots: Optional[List[Any]] = [], status: Optional[str] = '', techStatus: Optional[str] = '', type: Optional[str] = '', vmid: Optional[int] = None, **kwargs):
        self.id = id
        self.grid_id = gid
        self.name = name
        self.account_id = accountId
        self.account_name = accountName
        self.boot_partition = bootPartition
        self.created_time = createdTime
        self.deleted_time = deletedTime
        self.description = desc
        self.destruction_time = destructionTime
        self.device_name = devicename
        self.disk_path = diskPath
        self.guid = guid
        self.image_id = imageId
        self.images = images
        self.iotune = IOtune(**iotune) if isinstance(iotune, dict) else iotune
        self.milestones = milestones
        self.order = order
        self.params = params
        self.parent_id = parentId
        self.pci_slot = pciSlot
        self.pool = pool
        self.purge_attempts = purgeAttempts
        self.purge_time = purgeTime
        self.reality_device_number = realityDeviceNumber
        self.reference_id = referenceId
        self.res_id = resId
        self.res_name = resName
        self.role = role
        self.sep_id = sepId
        self.sep_type = sepType
        self.size_max = sizeMax
        self.size_used = sizeUsed
        self.snapshots = snapshots
        self.status = status
        self.tech_status = techStatus
        self.type = type
        self.vmid = vmid
        self.extra = Extra(kwargs)


import requests
import logging

logging.basicConfig(level=logging.CRITICAL)

from unittest import TestCase, mock
from requests.exceptions import RequestException
from pydigitalenergy.models import Result
from pydigitalenergy.adapter import RestAdapter
from pydigitalenergy.exceptions import DigitalEnergyApiException


class TestRestAdapter(TestCase):

    required_test_settings = {
        'hostname': 'localhost',
        'client_id': 'test',
        'client_secret': 'test',
        'token': 'test-jwt-key',
    } 

    def setUp(self) -> None:
        self.rest_adapter = RestAdapter(**self.required_test_settings)
        self.response = requests.Response()

    def test__make_request_good_request_returns_result(self):
        self.response.status_code = 200
        self.response._content = '{}'.encode()
        with mock.patch('requests.request', return_value=self.response):
            result = self.rest_adapter._make_request('GET', '')
            self.assertIsInstance(result, Result)

    def test__make_request_bad_request_raises_digitalenergyapi_exception(self):
        with mock.patch('requests.request', side_effect=RequestException):
            with self.assertRaises(DigitalEnergyApiException):
                self.rest_adapter._make_request('GET', '')

    def test__make_request_bad_json_raises_digitalenergyapi_exception(self):
        self.response._content = '{"some bad json": '
        with mock.patch('requests.request', return_value=self.response):
            with self.assertRaises(DigitalEnergyApiException):
                self.rest_adapter._make_request('GET', '')

    def test__make_request_300_or_higher_raises_digitalenergyapi_exception(self):
        self.response.status_code = 300
        with mock.patch('requests.request', return_value=self.response):
            with self.assertRaises(DigitalEnergyApiException):
                self.rest_adapter._make_request('GET', '')

    def test__make_request_199_or_lower_raises_digitalenergyapi_exception(self):
        self.response.status_code = 199
        with mock.patch('requests.request', return_value=self.response):
            with self.assertRaises(DigitalEnergyApiException):
                self.rest_adapter._make_request('GET', '')

    def test_get_method_passes_in_get(self):
        self.response.status_code = 200
        self.response._content = '{}'.encode()
        with mock.patch('requests.request', return_value=self.response) as request:
            self.rest_adapter.get('')
            self.assertTrue(request.method, 'GET')



from typing import Union, Optional
from pydigitalenergy.models import Extra


class Stats:
    def __init__(self, cpu: int, disksize: int, extips: int, exttraffic: int, gpu: int, ram: int, **kwargs):
        self.cpu = cpu
        self.disksize = disksize
        self.extips = extips
        self.exttraffic = exttraffic
        self.gpu = gpu
        self.ram = ram
        self.extra = Extra(kwargs)


class Resource:
    def __init__(self, Current: Union[Stats, dict], Reserved: Union[Stats, dict], **kwargs):
        self.current = Stats(**Current) if isinstance(Current, dict) else Current
        self.reserved = Stats(**Reserved) if isinstance(Reserved, dict) else Reserved
        self.extra = Extra(kwargs)


class Grid:
    def __init__(self, id: int, name: str, gid: int, Resources: Union[Resource, dict], flag: Optional[str] = '', guid: Optional[int] = None, locationCode: Optional[str] = '', **kwargs):
        self.id = id
        self.name = name
        self.data_center_id = gid
        self.resources = Resource(**Resources) if isinstance(Resources, dict) else Resources
        self.flag = flag
        self.guid = guid
        self.location_code = locationCode
        self.extra = Extra(kwargs)



from copy import deepcopy
from typing import List, Dict
from pydigitalenergy.models import Result
from pydigitalenergy.cloudbroker import models
from pydigitalenergy.cloudbroker.endpoints import API_PATH
from pydigitalenergy.adapter import RestAdapter
from pydigitalenergy.basegroup import BaseGroup
from pydigitalenergy.exceptions import DigitalEnergyApiException


class Cloudbroker:
    def __init__(self, adapter: RestAdapter):
        self.computes = ComputeGroup(adapter)
        self.vms = VMGroup(adapter)
        self.stacks = StackGroup(adapter)
        self.nodes = NodeGroup(adapter)
        self.hosts = HostGroup(adapter)
        self.clusters = ClusterGroup(adapter)
        self.grids = GridGroup(adapter)
        self.data_centers = DataCenterGroup(adapter)
        self.storages = StorageEndpointGroup(adapter)
        self.images = IamgeGroup(adapter)
        self.disks = DiskGroup(adapter)
        self.networks = ExternalNetworkGroup(adapter)


class ComputeGroup(BaseGroup):
    """
    Compute Group
    """

    def list(self, include_deleted: bool = False, page: int = 0, size: int = 0) -> List[models.Compute]:
        """
        Get list of compute instances
        Page and size can be specified only together

        :param include_deleted: (optional)
            Whether add deleted compute instances or not
        :param page: (optional)
            Page number
        :param size: (optional)
            Amount of instances per page
        
        :return: 
            List of Compute objects
        """
        result = self._get_data(API_PATH['compute_list'], {'includedeleted': include_deleted, 'page': page, 'size': size})
        return [models.Compute(**compute) for compute in result.data]

    def get(self, compute_id: int, reason: str = '') -> models.Compute:
        """
        Get information about compute instance

        :param compute_id: 
            Compute instance identifier
        :param reason: (optional)
            Reason of action
        
        :return: 
            Compute object
        """
        result = self._get_data(API_PATH['compute_get'], {'computeId': compute_id, 'reason': reason})
        return models.Compute(**result.data)


class StackGroup(BaseGroup):
    """
    Stack Group
    """

    def list(self, enabled: bool = True, page: int = 0, size: int = 0) -> List[models.Stack]:
        """
        Get list of stack instances
        Page and size can be specified only together

        :param enabled: (optional)
            Whether to show only enabled stacks or all of them
        :param page: (optional)
            Page number
        :param size: (optional)
            Amount of instances per page
        
        :return: 
            List of Stack objects
        """
        result = self._get_data(API_PATH['stack_list'], {'enabled': enabled, 'page': page, 'size': size})
        return [models.Stack(**stack) for stack in result.data]

    def get(self, stack_id: int) -> models.Stack:
        """
        Get information about stack instance

        :param stack_id: 
            Stack instance identifier

        :return: 
            Stack object
        """
        result = self._get_data(API_PATH['stack_get'], {'stackId': stack_id})
        return models.Stack(**result.data)


class NodeGroup(BaseGroup):
    """
    Node Group
    """

    def list(self, page: int = 0, size: int = 0) -> List[models.Node]:
        """
        Get list of node instances
        Page and size can be specified only together

        :param page: (optional)
            Page number
        :param size: (optional)
            Amount of instances per page
        
        :return: 
            List of Node objects
        """
        result = self._get_data(API_PATH['node_list'], {'page': page, 'size': size})
        return [models.Node(**node) for node in result.data]

    def get(self, node_id: int) -> models.Node:
        """
        Get information about node instance

        :param node_id: 
            Node instance identifier

        :return: 
            Node object
        """
        result = self._get_data(API_PATH['node_get'], {'nid': node_id})
        return models.Node(**result.data)


class BaseClusterGroup(BaseGroup):
    """
    
    """

    def _prep_stack_for_cluster(self, data: Dict):
        self._remove_fields(data, ('apiUrl', 'apikey'))

    def _prep_node(self, data: Dict):
        self._remove_fields(data, ('gid', 'id', 'name', 'status'))

    def _find_node(self, nodes: Result, stack: Dict) -> Dict:
        if stack is not None:
            for node in nodes.data:
                node_id = int(stack['referenceId']) if stack['referenceId'].isdigit() else stack['referenceId']
                if node_id == node.get('id', None):
                    return deepcopy(node)
        return None


class ClusterGroup(BaseClusterGroup):
    """
    Cluster Group
    """

    def list(self, page: int = 0, size: int = 0) -> List[models.Cluster]:
        """
        Get list of cluster instances
        Page and size can be specified only together

        :param page: (optional)
            Page number
        :param size: (optional)
            Amount of instances per page
        
        :return: 
            List of Cluster objects
        """
        stacks = self._get_data(API_PATH['stack_list'], {'page': page, 'size': size})
        nodes = self._get_data(API_PATH['node_list'], {'page': page, 'size': size})

        clusters = []
        for stack in stacks.data:
            node = self._find_node(nodes, stack)
            
            if node:
                self._prep_stack_for_cluster(stack)
                self._prep_node(node)
                clusters.append(models.Cluster(**{**stack, **node}))

        return clusters

    def get(self, cluster_id: int) -> models.Cluster:
        """
        Get information about cluster instance

        :param cluster_id: 
            Cluster instance identifier

        :return: 
            Cluster object
        """
        stack = self._get_data(API_PATH['stack_get'], {'stackId': cluster_id})
        self._prep_stack_for_cluster(stack.data)

        node = self._get_data(API_PATH['node_get'], {'nid': stack.data['referenceId']})
        self._prep_node(node.data)

        return models.Cluster(**{**stack.data, **node.data})


class BaseHostGroup(BaseGroup):
    """
    
    """

    def _prep_stack_for_host(self, data: Dict):
        self._remove_fields(data, ('gid', 'name', 'status', 'desc', 'eco'))
        self._rename_fields(data, {'id': 'stackId'})

    def _prep_sep(self, data: Dict):
        self._remove_fields(data, ('id', 'gid', 'guid'))
        self._rename_fields(data, {'name': 'sepName', 'type': 'sepType'})

    def _find_stack(self, stacks: Result, node: Dict) -> Dict:
        if node is not None:
            for stack in stacks.data:
                node_id = int(stack['referenceId']) if stack['referenceId'].isdigit() else stack['referenceId']
                if node_id == node.get('id', None):
                    return deepcopy(stack)
        return None

    def _find_sep(self, seps: Result, node: Dict) -> Dict:
        if node is not None:
            for sep in seps.data:
                if node.get('id', None) in sep['consumedBy']:
                    return deepcopy(sep)
        return None


class HostGroup(BaseHostGroup):
    """
    Host Group
    """

    def list(self, enabled: bool = True, page: int = 0, size: int = 0) -> List[models.Host]:
        """
        Get list of host instances
        Page and size can be specified only together

        :param enabled: (optional)
            Whether to show only enabled hosts or all of them
        :param page: (optional)
            Page number
        :param size: (optional)
            Amount of instances per page
        
        :return: 
            List of Host objects
        """
        nodes = self._get_data(API_PATH['node_list'], {'page': page, 'size': size})
        stacks = self._get_data(API_PATH['stack_list'], {'enabled': enabled, 'page': page, 'size': size})
        seps = self._get_data(API_PATH['sep_list'], {'page': page, 'size': size})

        hosts = []
        for node in nodes.data:
            stack = self._find_stack(stacks, node)
            sep = self._find_sep(seps, node)

            if stack and sep:
                self._prep_stack_for_host(stack)
                self._prep_sep(sep)
                hosts.append(models.Host(**{**node, **stack, **sep}))

        return hosts

    def get(self, host_id: int) -> models.Host:
        """
        Get information about host instance

        :param host_id: 
            Host instance identifier

        :return: 
            Host object
        """
        node = self._get_data(API_PATH['node_get'], {'nid': host_id})
        stacks = self._get_data(API_PATH['stack_list'])
        seps = self._get_data(API_PATH['sep_list'])

        stack = self._find_stack(stacks, node.data)
        sep = self._find_sep(seps, node.data)

        if stack and sep:
            self._prep_stack_for_host(stack)
            self._prep_sep(sep)
            return models.Host(**{**node.data, **stack, **sep})

        raise DigitalEnergyApiException('404: Not Found')


class BaseVMGroup(BaseGroup):
    """
    
    """

    def _find_stack_for_vm(self, stacks: Result, compute: Dict) -> Dict:
        if compute is not None:
            for stack in stacks.data:
                if stack['id'] == compute.get('stackId', None):
                    return deepcopy(stack)
            return None

    def _find_image(self, images: Result, compute: Dict) -> Dict:
        if compute is not None:
            for image in images.data:
                if image['id'] == compute.get('imageId', None):
                    return deepcopy(image)
            return {'hasImage': False}


class VMGroup(BaseVMGroup, BaseHostGroup, BaseClusterGroup):
    """
    Virtual Machine Group
    """

    def list(self, include_deleted: bool = False, page: int = 0, size: int = 0) -> List[models.VM]:
        """
        Get list of virtual machine instances
        Page and size can be specified only together

        :param include_deleted: (optional)
            Whether add deleted compute instances or not
        :param page: (optional)
            Page number
        :param size: (optional)
            Amount of instances per page
        
        :return: 
            List of Virtual Machine objects
        """
        computes = self._get_data(API_PATH['compute_list'], {'includedeleted': include_deleted, 'page': page, 'size': size})
        nodes = self._get_data(API_PATH['node_list'], {'page': page, 'size': size})
        stacks = self._get_data(API_PATH['stack_list'], {'page': page, 'size': size})
        seps = self._get_data(API_PATH['sep_list'], {'page': page, 'size': size})
        images = self._get_data(API_PATH['image_list'], {'page': page, 'size': size})

        vms = []
        for compute in computes.data:

            # Host
            host_stack = self._find_stack_for_vm(stacks, compute)
            host_node = self._find_node(nodes, host_stack)
            host_sep = self._find_sep(seps, host_node)

            if host_stack and host_node and host_sep:
                self._prep_stack_for_host(host_stack)
                self._prep_sep(host_sep)
                self._add_fields_prefix(host_stack, 'host')
                self._add_fields_prefix(host_node, 'host')
                self._add_fields_prefix(host_sep, 'host')

            # Cluster
            cluster_stack = self._find_stack_for_vm(stacks, compute)
            cluster_node = self._find_node(nodes, cluster_stack)

            if cluster_stack and cluster_node:
                self._prep_stack_for_cluster(cluster_stack)
                self._prep_node(cluster_node)
                self._add_fields_prefix(cluster_stack, 'cluster')
                self._add_fields_prefix(cluster_node, 'cluster')

            # Image
            image = self._find_image(images, compute)
            self._add_fields_prefix(image, 'image')

            # Virtual machine
            if host_stack and host_node and host_sep and cluster_stack and cluster_node:
                vms.append(models.VM(**{
                    **compute,
                    **host_stack, **host_node, **host_sep,
                    **cluster_stack, **cluster_node,
                    **image
                }))

        return vms

    def get(self, vm_id: int, reason: str = '') -> models.VM:
        """
        Get information about Virtual Machine instance

        :param vm_id: 
            Virtual Machine instance identifier
        :param reason: (optional)
            Reason of action
        
        :return: 
            Virtual Machine object
        """

        compute = self._get_data(API_PATH['compute_get'], {'computeId': vm_id, 'reason': reason})
        stack = self._get_data(API_PATH['stack_get'], {'stackId': compute.data['stackId']})
        images = images = self._get_data(API_PATH['image_list'])
        nodes = self._get_data(API_PATH['node_list'])
        seps = self._get_data(API_PATH['sep_list'])

        # Host
        host_stack = deepcopy(stack.data)
        host_node = self._find_node(nodes, host_stack)
        host_sep = self._find_sep(seps, host_node)

        if host_stack and host_node and host_sep:
            self._prep_stack_for_host(host_stack)
            self._prep_sep(host_sep)
            self._add_fields_prefix(host_stack, 'host')
            self._add_fields_prefix(host_node, 'host')
            self._add_fields_prefix(host_sep, 'host')

        # Cluster
        cluster_stack = deepcopy(stack.data)
        cluster_node = self._find_node(nodes, cluster_stack)

        if cluster_stack and cluster_node:
            self._prep_stack_for_cluster(cluster_stack)
            self._prep_node(cluster_node)
            self._add_fields_prefix(cluster_stack, 'cluster')
            self._add_fields_prefix(cluster_node, 'cluster')

        # Image
        image = self._find_image(images, compute.data)
        self._add_fields_prefix(image, 'image')

        # Virtual machine
        if host_stack and host_node and host_sep and cluster_stack and cluster_node:
            return models.VM(**{
                **compute.data,
                **host_stack, **host_node, **host_sep,
                **cluster_stack, **cluster_node,
                **image
            })
        else:
            details = "Compute without Host and Cluster doesn't appear in Virtual Machine actor."
            raise DigitalEnergyApiException('404: Not Found. ' + details)


class GridGroup(BaseGroup):
    """
    Grid Group
    """

    def list(self, page: int = 0, size: int = 0) -> List[models.Grid]:
        """
        Get list of grid instances
        Page and size can be specified only together

        :param page: (optional)
            Page number
        :param size: (optional)
            Amount of instances per page
        
        :return: 
            List of Grid objects
        """
        result = self._get_data(API_PATH['grid_list'], {'page': page, 'size': size})
        return [models.Grid(**grid) for grid in result.data]

    def get(self, grid_id: int) -> models.Grid:
        """
        Get information about grid instance

        :param grid_id: 
            Grid instance identifier

        :return: 
            Grid object
        """
        result = self._get_data(API_PATH['grid_get'], {'gridId': grid_id})
        if isinstance(result.data, dict):
            return models.Grid(**result.data)
        else:
            raise DigitalEnergyApiException('404: Not Found. ' + result.data if isinstance(result.data, str) else '')


class DataCenterGroup(GridGroup):
    """
    Data Center Group 
    It is an alias for Grid Group
    """
    pass


class StorageEndpointGroup(BaseGroup):
    """
    Storage Endpoint Group
    """

    def list(self, page: int = 0, size: int = 0) -> List[models.StorageEndpoint]:
        """
        Get list of Storage Endpoint instances
        Page and size can be specified only together

        :param page: (optional)
            Page number
        :param size: (optional)
            Amount of instances per page
        
        :return: 
            List of StorageEndpoint objects
        """
        result = self._get_data(API_PATH['sep_list'], {'page': page, 'size': size})
        return [models.StorageEndpoint(**sep) for sep in result.data]

    def get(self, sep_id: int) -> models.StorageEndpoint:
        """
        Get information about Storage Endpoint instance

        :param sep_id: 
            Storage Endpoint instance identifier

        :return: 
            StorageEndpoint object
        """
        result = self._get_data(API_PATH['sep_get'], {'sep_id': sep_id})
        return models.StorageEndpoint(**result.data)

    def get_config(self, sep_id: int) -> models.StorageConfig:
        """
        Get config of Storage Endpoint instance

        :param sep_id: 
            Storage Endpoint instance identifier

        :return: 
            StorageConfig object
        """
        result = self._get_data(API_PATH['sep_get_config'], {'sep_id': sep_id})
        return models.StorageConfig(**result.data)

    def get_pool(self, sep_id: int, pool_name: str) -> models.StoragePool:
        """
        Get pool of Storage Endpoint instance

        :param sep_id: 
            Storage Endpoint instance identifier
        :param pool_name: 
            Name of certain Storage Endpoint pool

        :return: 
            StoragePool object
        """
        result = self._get_data(API_PATH['sep_get_pool'], {'sep_id': sep_id, 'pool_name': pool_name})
        return models.StoragePool(**result.data)


class IamgeGroup(BaseGroup):
    """
    Image Group
    """

    def list(self, sep_id: int = None, shared_with: int = None, page: int = 0, size: int = 0) -> List[models.Image]:
        """
        Get list of Image instances
        Page and size can be specified only together

        :param sep_id: (optional)
            Filter images based on Storage Endpoint instance identifier
        :param shared_with: (optional)
            Filter images based on Account identifier availability
        :param page: (optional)
            Page number
        :param size: (optional)
            Amount of instances per page
        
        :return: 
            List of Image objects
        """
        result = self._get_data(API_PATH['image_list'], {'sepId': sep_id, 'sharedWith': shared_with, 'page': page, 'size': size})
        return [models.Image(**image) for image in result.data]

    def get(self, image_id: int) -> models.Image:
        """
        Get information about Image instance

        :param image_id: 
            Image instance identifier

        :return: 
            Image object
        """
        result = self._get_data(API_PATH['image_get'], {'imageId': image_id})
        return models.Image(**result.data)


class DiskGroup(BaseGroup):
    """
    Disk Group
    """

    def list(self, account_id: int = None, disk_type: str = None, page: int = 0, size: int = 0) -> List[models.Disk]:
        """
        Get list of Disk instances
        Page and size can be specified only together

        :param account_id: (optional)
            Filter disks based on their belonging to the Account
        :param disk_type: (optional)
            Filter disks based on their type mark
        :param page: (optional)
            Page number
        :param size: (optional)
            Amount of instances per page
        
        :return: 
            List of Disk objects
        """
        result = self._get_data(API_PATH['disks_list'], {'accountId': account_id, 'type': disk_type, 'page': page, 'size': size})
        return [models.Disk(**disk) for disk in result.data]

    def get(self, disk_id: int) -> models.Disk:
        """
        Get information about Disk instance

        :param disk_id: 
            Disk instance identifier

        :return: 
            Disk object
        """
        result = self._get_data(API_PATH['disks_get'], {'diskId': disk_id})
        return models.Disk(**result.data)


class ExternalNetworkGroup(BaseGroup):
    """
    External Network Group
    """

    def list(self, account_id: int = None, page: int = 0, size: int = 0) -> List[models.ExternalNetwork]:
        """
        Get list of External Network instances
        Page and size can be specified only together

        :param account_id: (optional)
            Filter external networks based on their belonging to the Account
        :param page: (optional)
            Page number
        :param size: (optional)
            Amount of instances per page
        
        :return: 
            List of ExternalNetwork objects
        """
        result = self._get_data(API_PATH['extnet_list'], {'accountId': account_id, 'page': page, 'size': size})
        return [models.ExternalNetwork(**sep) for sep in result.data]

    def get(self, net_id: int) -> models.ExternalNetwork:
        """
        Get information about External Network instance

        :param net_id: 
            External Network instance identifier

        :return: 
            ExternalNetwork object
        """
        result = self._get_data(API_PATH['extnet_get'], {'net_id': net_id})
        return models.ExternalNetwork(**result.data)



from typing import List, Dict
from requests.structures import CaseInsensitiveDict


class Result:
    def __init__(self, status_code: int, headers: CaseInsensitiveDict, message: str = '', data: List[Dict] = None):
        """
        Result returned from low-level RestAdapter

        :param status_code: 
            Standard HTTP Status code
        :param message: 
            Human readable result
        :param data: 
            Python List of Dictionaries (or maybe just a single Dictionary on error)
        """
        self.status_code = int(status_code)
        self.headers = headers
        self.message = str(message)
        self.data = data if data else []


class Extra:
    def __init__(self, kwargs):
        """
        Extra fields for the strong models
        
        :param kwargs: 
            Dictionary with additional information
        """
        self.__dict__.update(kwargs)



from uuid import UUID
from typing import Union, Optional, List, Any
from pydigitalenergy.cloudbroker.models.disk import Disk
from pydigitalenergy.models import Extra


class Machine:
    """
    There is no example or documentation of Machine instance right now
    Therefore no strong model
    """
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)



from .compute import Compute
from .stack import Stack
from .node import Node
from .grid import Grid
from .sep import StorageEndpoint, StorageConfig, StoragePool
from .image import Image
from .disk import Disk
from .extnet import ExternalNetwork
from .cluster import Cluster
from .host import Host
from .vm import VM


from typing import Optional, Union, List, Any
from pydigitalenergy.cloudbroker.models.node import Consumption
from pydigitalenergy.cloudbroker.models.sep import StorageConfig
from pydigitalenergy.models import Extra


class Hardware:
    def __init__(self, sepName: str, desc: str, sepType: str, objStatus: str, techStatus: str, config: Optional[Union[StorageConfig, dict]] = None, consumedBy: Optional[List[int]] = [], milestones: Optional[int] = None, providedBy: Optional[List[Any]] = []):
        self.name = sepName
        self.description = desc
        self.type = sepType
        self.obj_status = objStatus
        self.tech_status = techStatus
        self.config = StorageConfig(**config) if isinstance(config, dict) else config
        self.consumed_by = consumedBy
        self.milestones = milestones
        self.provided_by = providedBy


class Host:
    def __init__(self, id: int, gid: int, name: str, status: str, type: str, stackId: int, sepName: str, desc: str, sepType: str, objStatus: str, techStatus: str, version: Optional[str], referenceId: int, apiUrl: Optional[str] = '', apikey: Optional[str] = '', descr: Optional[str] = '', drivers: Optional[List[str]] = '', error: Optional[int] = None, guid: Optional[int] = None, ipaddr: Optional[List[str]] = [], consumption: Union[Consumption, dict] = None, roles: Optional[List[str]] = [], config: Optional[Union[StorageConfig, dict]] = None, consumedBy: Optional[List[int]] = [], milestones: Optional[int] = None, providedBy: Optional[List[Any]] = [], **kwargs):
        self.id = id
        self.data_center_id = gid
        self.name = name
        self.status = status
        self.version = version
        self.ip_address = ipaddr
        self.consumption = Consumption(**consumption) if isinstance(consumption, dict) else consumption
        self.roles = roles
        self.hardware = Hardware(sepName, desc, sepType, objStatus, techStatus, config, consumedBy, milestones, providedBy)
        self.cluster_id = stackId
        self.type = type
        self.node_id = int(referenceId) if referenceId.isdigit() else referenceId
        self.description = descr
        self.drivers = drivers
        self.api_url = apiUrl
        self.api_key = apikey
        self.error = error
        self.guid = guid
        self.extra = Extra(kwargs)



from typing import Optional, List, Union
from pydigitalenergy.cloudbroker.models.node import Consumption
from pydigitalenergy.models import Extra


class Cluster:
    def __init__(self, id: int, gid: int, name: str, status: str, referenceId: int, descr: Optional[str] = '', error: Optional[int] = None, guid: Optional[int] = None, consumption: Union[Consumption, dict] = None, roles: Optional[List[str]] = [], **kwargs):
        self.id = id
        self.node_id = int(referenceId) if referenceId.isdigit() else referenceId
        self.data_center_id = gid
        self.guid = guid
        self.name = name
        self.status = status
        self.roles = roles
        self.description = descr
        self.error = error
        self.consumption = Consumption(**consumption) if isinstance(consumption, dict) else consumption
        self.extra = Extra(kwargs)



import requests
import logging

from typing import Dict
from json import JSONDecodeError
from pydigitalenergy.exceptions import DigitalEnergyApiException
from pydigitalenergy.models import Result


class RestAdapter:
    def __init__(self, hostname: str, client_id: str, client_secret: str, auth_version: str = 'v1', auth_validity: int = 3600, ssl_verify: bool = True, logger: logging.Logger = None, token: str = ''):
        """
        Constructor for RestAdapter

        :param hostname: 
            DigitalEnergy API hostname. Something like, api.digitalenergy.online
        :param client_id: 
            String defined application identifier
        :param client_secret: 
            String defined application secret
        :param auth_version: 
            Auth version, default v1
        :param auth_validity: 
            Lifetime of JWT-key in seconds, default 3600 seconds or 1 hour
        :param ssl_verify: 
            Normally set to True, but if having SSL/TLS cert validation issues, can turn off with False
        :param logger: (optional)
            If your app has a logger, pass it in here
        :param token: (optional)
            Your JWT-key to work with API, if not specified will be generated automatically
        """
        if not ssl_verify:
            requests.packages.urllib3.disable_warnings()

        self.url = f'https://{hostname}/restmachine'
        self._hostname = hostname
        self._client_id = client_id
        self._client_secret = client_secret
        self._ssl_verify = ssl_verify
        self._logger = logger or logging.getLogger(__name__)
        self._token = token if token else self._get_new_token(auth_version, auth_validity)

    def _get_new_token(self, auth_version: str = 'v1', auth_validity: int = 3600) -> str:
        """
        Private method for getting new JSON Web Token (JWT)

        :param auth_version: 
            Auth version, default v1
        :param auth_validity: 
            Lifetime of JWT in seconds, default 3600 seconds or 1 hour
        
        :return: 
            Token string
        """
        auth_params = {
            'client_id': self._client_id,
            'client_secret': self._client_secret, 
            'validity': auth_validity,
            'grant_type': 'client_credentials', 
            'response_type': 'id_token'
        }
        auth_url = f'https://sso-{self._hostname}/{auth_version}/oauth/access_token'
        log_line_pre = f'method=POST, url={auth_url}'
        log_line_post = ', '.join((log_line_pre, 'success={}, status_code={}, message={}'))

        # Log HTTP params and perform an HTTP POST request, catching and re-raising any exceptions
        try:
            self._logger.debug(msg=log_line_pre)
            response = requests.post(url=auth_url, verify=self._ssl_verify, params=auth_params)
        except requests.exceptions.RequestException as e:
            self._logger.error(msg=(str(e)))
            raise DigitalEnergyApiException('Request new JWT failed') from e

        # If status_code not in 200-299 range raise exception, otherwise return new JWT
        is_success = 299 >= response.status_code >= 200
        log_line = log_line_post.format(is_success, response.status_code, response.reason)

        if not is_success:
            self._logger.error(msg=log_line)
            raise DigitalEnergyApiException(f'{response.status_code}: {response.reason}')

        self._logger.debug(msg=log_line)
        return response.content.decode('utf8')

    def _make_request(self, http_method: str, endpoint: str, ep_params: Dict = None, data: Dict = None) -> Result:
        """
        Private method for get(), post(), delete(), etc. methods

        :param http_method: 
            GET, POST, DELETE, etc.
        :param endpoint: 
            URL Endpoint as a string
        :param ep_params: (Optional)
            Dictionary of Endpoint parameters
        :param data: (Optional)
            Dictionary of data to pass to DigitalEnergyApi
        
        :return: 
            Result object
        """
        headers = {
            'accept': 'application/json',
            'Authorization': f'bearer {self._token}',
            'Content-Type': 'application/x-www-form-urlencoded'
        }
        full_url = self.url + endpoint
        log_line_pre = f'method={http_method}, url={full_url}, params={ep_params}'
        log_line_post = ', success={}, status_code={}, message={}'

        # Log HTTP params and perform an HTTP request, catching and re-raising any exceptions
        try:
            self._logger.debug(msg=log_line_pre)
            response = requests.request(method=http_method, url=full_url, verify=self._ssl_verify,
                                        headers=headers, params=ep_params, json=data)
        except requests.exceptions.RequestException as e:
            self._logger.error(msg=(str(e)))
            raise DigitalEnergyApiException('Request failed') from e

        # Deserialize JSON output to Python object, or return failed Result on exception
        try:
            data_out = response.json()
        except (ValueError, TypeError, JSONDecodeError) as e:
            self._logger.error(msg=log_line_pre + log_line_post.format(False, None, e))
            raise DigitalEnergyApiException('Bad JSON in response') from e

        # If status_code in 200-299 range, return success Result with data, otherwise raise exception
        is_success = 299 >= response.status_code >= 200
        log_line = log_line_pre + log_line_post.format(is_success, response.status_code, response.reason)

        if is_success:
            self._logger.debug(msg=log_line)
            return Result(response.status_code, headers=response.headers, message=response.reason, data=data_out)

        self._logger.error(msg=log_line)
        raise DigitalEnergyApiException(f'{response.status_code}: {response.reason}')

    def get(self, endpoint: str, ep_params: Dict = None) -> Result:
        """
        HTTP GET method

        :param endpoint: 
            URL Endpoint as a string
        :param ep_params: (Optional)
            Dictionary of Endpoint parameters
        :param data: (Optional)
            Dictionary of data to pass to DigitalEnergyApi
        
        :return: 
            Result object
        """
        return self._make_request(http_method='GET', endpoint=endpoint, ep_params=ep_params)

    def post(self, endpoint: str, ep_params: Dict = None, data: Dict = None) -> Result:
        """
        HTTP POST method

        :param endpoint: 
            URL Endpoint as a string
        :param ep_params: (Optional)
            Dictionary of Endpoint parameters
        :param data: (Optional)
            Dictionary of data to pass to DigitalEnergyApi
        
        :return: 
            Result object
        """
        return self._make_request(http_method='POST', endpoint=endpoint, ep_params=ep_params, data=data)
